/**
 @file  test_demux_mxf_ext_io.cpp
 @brief command-line sample for MXF demuxer

 @verbatim
 File: test_demux_mxf_ext_io.cpp

 Desc: command-line sample for MXF demuxer

 Copyright (c) 2021 MainConcept GmbH or its affiliates. All rights reserved.

 MainConcept and its logos are registered trademarks of MainConcept GmbH or its affiliates.
 This software is protected by copyright law and international treaties. Unauthorized
 reproduction or distribution of any portion is prohibited by law.
 @endverbatim
 **/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <memory>
#if defined(_WIN32) || defined(_WIN64)
#include <Windows.h>
#else
#include <unistd.h>
#endif
#if defined(__APPLE__)
#include <mach/mach_time.h>
#else
#include <ctime>
#endif

#include "demux_mxf.h"
#include "auxinfo.h"
#include "buf_file.h"
#include "mccolorimetry.h"

#include "sample_common_misc.h"

 //#define SHOW_AUX_INFO // catch demuxer auxinfo and print timecode and chunk info

#define MAX_STREAMS		64	// the max streams we will handle

struct drive_struct;
struct global_vars_struct;
class CExtIO : public mc_external_io_t
{
#if !defined(__APPLE__) && !defined(__linux__)
    static int fseek__(FILE* __stream, int64_t __off, int __whence)
    {
        return ::_fseeki64(__stream, __off, __whence);
    }
    static int64_t ftell__(FILE* __stream)
    {
        return ::_ftelli64(__stream);
    }
#else
    static int fseek__(FILE* __stream, int64_t __off, int __whence)
    {
        return ::fseek(__stream, __off, __whence);
    }
    static int64_t ftell__(FILE* __stream)
    {
        return ::ftell(__stream);
    }
#endif

    size_t m_file_pos;
    FILE *m_log_file;

    int32_t Open(const char *fileName, void **pp_file_handle)
    {
        *pp_file_handle = 0;
        FILE *fd = fopen(fileName, "rb");
        if (fd)
        {
            *pp_file_handle = fd;
        }
        fprintf(m_log_file, "ExtIO: Open file:'%s' FILE*:0x%p\n", fileName, *pp_file_handle);
        fflush(m_log_file);
        return fd != 0 ? 0 : 1;
    }
    int32_t Seek(void *p_file_handle, int64_t position)
    {
        FILE *fd = (FILE *)p_file_handle;
        m_file_pos = position;
        if (fd)
        {
            return fseek__(fd, position, SEEK_SET) == 0 ? 0 : 1;
        }
        return 1;
    }
    int32_t Read(void *p_file_handle, uint8_t *p_buffer, int32_t buffer_size)
    {
        FILE *fd = (FILE *)p_file_handle;
        if (m_log_file)
        {
            fprintf(m_log_file, "ExtIO: pos: %" PRIu64 ", read: %d\n", m_file_pos, buffer_size);
            fflush(m_log_file);
        }
        if (fd)
        {
            int32_t bytes = (int32_t)fread(p_buffer, 1, buffer_size, fd);
            if (bytes > 0)
                return bytes;
        }
        return 0;
    }
    void Close(void **pp_file_handle)
    {
        fprintf(m_log_file, "ExtIO: Close FILE*:0x%p\n", *pp_file_handle);
        fflush(m_log_file);
        if (*pp_file_handle)
            fclose((FILE *)*pp_file_handle);
        *pp_file_handle = 0;
    }

    static int32_t mxfExternalOpenCB(void *p_app_ptr, const char *p_filename, void **pp_file_handle)
    {
        return ((CExtIO *)p_app_ptr)->Open(p_filename, pp_file_handle);
    }

    static int32_t mxfExternalSeekCB(void *p_app_ptr, void *p_file_handle, int64_t position)
    {
        return ((CExtIO *)p_app_ptr)->Seek(p_file_handle, position);
    }

    static int32_t mxfExternalReadCB(void *p_app_ptr, void *p_file_handle, uint8_t *p_buffer, int32_t buffer_size)
    {
        return ((CExtIO *)p_app_ptr)->Read(p_file_handle, p_buffer, buffer_size);
    }

    static void mxfExternalCloseCB(void *p_app_ptr, void **pp_file_handle)
    {
        ((CExtIO *)p_app_ptr)->Close(pp_file_handle);
    }

public:
    CExtIO(const char *input_file_name, mxfdmux_settings &demuxer_set, const char *log_file_name)
        : m_log_file(0), m_file_pos(0)
    {
        file_size = GetFileSize(input_file_name);
        p_app_ptr = this;
        p_open = mxfExternalOpenCB;
        p_openW = 0;
        p_seek = mxfExternalSeekCB;
        p_read = mxfExternalReadCB;
        p_close = mxfExternalCloseCB;

        demuxer_set.use_external_io = 1;
        demuxer_set.p_external_io = this;

        if (log_file_name)
        {
            m_log_file = fopen(log_file_name, "w");
        }
    }
    int64_t GetFileSize(const char *input_file_name) 
    {
        int64_t fileSize{};
        FILE *fd = fopen(input_file_name, "rb");
        if (fd)
        {
            fseek__(fd, 0, SEEK_END);
            fileSize = ftell__(fd);
            fclose(fd);
        }
        return fileSize;
    }
};

enum MXF_SEEK_RESULT
{
    MXF_SEEK_SUCCESSFULL = 0,
    MXF_SEEK_EOF = 1,
    MXF_SEEK_ERROR
};
struct stream_info_struct
{
    mxfdmux_stream_info info;
    mxfdmux_stream_settings set;
    mxfdmux_stream_format_tt stream_format;

    int64_t total_demuxed;

    global_vars_struct *vars;
    int idx;
};


struct stream_to_get_struct
{
    int32_t ID;
    char path[2048];
};

struct global_vars_struct
{
    mxfdmux_tt *demuxer{};
    mxfdmux_settings dmux_set{};
    mxfdmux_file_info file_info{};
    mxfdmux_parser_settings parser_set{};
    int32_t parser{};

    char input_file[2048]{};

    //int32_t dm_demo;
    uint16_t preview_range{};

    int32_t stream_count{};
    stream_info_struct streams[MAX_STREAMS]{};
    mxfdmux_dm_track_info_tt dm_tracks[MAX_STREAMS]{};

    // streams to demux
    int32_t num_streams_to_get{};
    stream_to_get_struct streams_to_get[MAX_STREAMS]{};

    int64_t prev_processed{};
    int64_t curr_processed{};
    int64_t seek_position_start{};
    int64_t seek_position_stop{-1};

    bool is_growing_file{};

    uint32_t gfile_open_timeout{};
    bool use_ext_io{};
    char ext_log_file[2048]{};

    bool print_auxinfo{};
};

static int detect_contents(global_vars_struct *vars)
{
    //int32_t i, j, k;

    mc_time_code_t  time_code;
    memset(&time_code, 0, sizeof(mc_time_code_t));

    printf("\nContainer contents:\n");

    if (0 == mxfDemuxGetStartTimecode(vars->demuxer, &time_code))
        printf("\n  Start timecode: %.02d:%.02d:%.02d:%.02d\n", time_code.Hour, time_code.Min, time_code.Sec, time_code.Frm);

    if (mxfDemuxGetFileInfo(vars->demuxer, &vars->file_info))
    {
        printf("Unable to get file info\n");
        return 1;
    }

    vars->stream_count = vars->file_info.stream_count;
    if (vars->stream_count > MAX_STREAMS)
        vars->stream_count = MAX_STREAMS;

    printf("\n  Essence streams: %d \n\n", vars->file_info.stream_count);
    for (int i = 0; i < vars->stream_count; i++)
    {
        if (mxfDemuxGetStreamFormat(vars->demuxer, &vars->streams[i].stream_format, i))
        {
            printf("Unable to get stream format for stream %d\n", i);
            return 1;
        }

        char stream_type[100] = "";

        switch (vars->streams[i].stream_format.format.stream_mediatype)
        {
            case mctMPEG2V:     sprintf(stream_type, "MPEG-2 video stream");    break;
            case mctMPEG4V:     sprintf(stream_type, "MPEG-4 video stream");    break;
            case mctDV:         sprintf(stream_type, "DV video stream");        break;
            case mctJ2K:        sprintf(stream_type, "JPEG2000 video stream");  break;
            case mctVC3:        sprintf(stream_type, "VC3 video stream");       break;
            case mctH264:       sprintf(stream_type, "AVC video stream");       break;
            case mctAES3_331M:  sprintf(stream_type, "AES 331M audio stream");  break;
            case mctAES3_382M:  sprintf(stream_type, "AES 382M audio stream");  break;
            case mctPCM:        sprintf(stream_type, "PCM audio stream");       break;
            case mctAIFF:       sprintf(stream_type, "AIFF audio stream");      break;
            case mctALAW:       sprintf(stream_type, "A-law audio stream");     break;
            case mctMPEG1A:     sprintf(stream_type, "MPEG-1 audio stream");    break;
            case mctVBI_Data:   sprintf(stream_type, "VBI data stream");        break;
            case mctANC_Data:   sprintf(stream_type, "ANC data stream");        break;
            case mctProRes:   sprintf(stream_type, "ProRes video stream");        break;
            case mctHEVC:   sprintf(stream_type, "HEVC video stream");        break;
            default:  sprintf(stream_type, "Unknown stream");
        }

        printf("    Stream ID %d : Track ID %d, %s, %llu samples", vars->streams[i].stream_format.ID, vars->streams[i].stream_format.track_id, stream_type, (unsigned long long)vars->streams[i].stream_format.sample_count);
        if (vars->streams[i].stream_format.format.stream_mediatype > mctMinVideoType && vars->streams[i].stream_format.format.stream_mediatype < mctMaxVideoType)
            printf(", AFD 0x%01X", vars->streams[i].stream_format.active_format_descriptor);
        if( vars->streams[i].stream_format.descriptor_type == GenericPicture && vars->streams[i].stream_format.descriptor ) {
            mxf_generic_picture_descriptor_t* desc = (mxf_generic_picture_descriptor_t*)vars->streams[i].stream_format.descriptor;
            printf("\n                          generic_picture_descriptor: display_height %d, display_width %d\n", desc->display_height, desc->display_width);
            printf("                          generic_picture_descriptor: stored_width %d, stored_height %d\n",   desc->stored_width,   desc->stored_height);
            printf("                          generic_picture_descriptor: sampled_width %d, sampled_height %d\n", desc->sampled_width, desc->sampled_height);
        }
        vars->streams[i].idx = i;
        printf("\n");
    }

    printf("\n");

    for (int i = 0; i < vars->stream_count; i++)
    {
        printf("    Stream ID %d : positionable: duration %lld, sample count %lld \n",
            vars->streams[i].stream_format.ID,
            (long long)vars->streams[i].stream_format.positionable_duration,
            (long long)vars->streams[i].stream_format.positionable_sample_count);
    }

    printf("\n");

    printf("\n  Descriptive Metadata Schemes: %d\n\n", vars->file_info.dm_schemes_count);
    for (int i = 0; i < vars->file_info.dm_schemes_count; i++)
    {
        char chUL[100] = "   ";
        for (int j = 0; j < 16; j++)
        {
            char chVal[4] = "";
            sprintf(chVal, " %02X", vars->file_info.dm_schemes[i][j]);
            strcat(chUL, chVal);
        }
        strcat(chUL, "\n");
        printf("%s", chUL);
    }

    printf("\n  Descriptive Metadata Framework sets: %d\n\n", vars->file_info.dm_framework_set_count);


    printf("\n  Descriptive Metadata tracks: %d\n", vars->file_info.dm_tracks_count);

    printf("\n");

    return 0;
}


static uint32_t demux_auxinfo(bufstream_tt *bs, uint32_t offs, uint32_t info_ID, void *info_ptr, uint32_t info_size)
{
    dmux_chunk_info *chunk_info;
    //struct impl_stream* p = bs->Buf_IO_struct;
    //uint64_t *ptr;
    mc_time_code_t *time_code_info;
    sample_struct *sample_info;

    stream_info_struct *stream_info = (stream_info_struct*)bs->drive_ptr;

    switch (info_ID)
    {
        // mcapi push mode implementation
        case OVERRIDE_NATIVE_COLORIMETRY:
        {
            colorimetry_t* colorimetry = (colorimetry_t*)info_ptr;
            if (colorimetry && (info_size == sizeof(colorimetry_t)))
            {
                printf("OVERRIDE_NATIVE_COLORIMETRY %d:\n", stream_info->idx);
                printf("		color_primaries %d\n", colorimetry->color_primaries);
                printf("		transfer_characteristics %d\n", colorimetry->transfer_characteristics);
                printf("		matrix_coefficients %d\n", colorimetry->matrix_coefficients);
                printf("\n");
            }
        }
            break;

        case DMUX_CHUNK_INFO:
            chunk_info = (dmux_chunk_info*)info_ptr;
            if (chunk_info && (info_size == sizeof(dmux_chunk_info)))
            {
                stream_info->vars->curr_processed = chunk_info->file_pos;
                stream_info->total_demuxed += chunk_info->length;
                if (stream_info->vars->print_auxinfo) {
                    printf("DMUX_CHUNK_INFO %d:\n", stream_info->idx);
                    // show AES-3 331 header bytes - only sent when mxfdmux_settings->deliver_aes3_331_headers = 1
                    if (chunk_info->info_len && (chunk_info->flags & MCDMUX_FLAGS_AES3_331M))
                    {
                        printf("		AES-3 331 header bytes = 0x%02X%02X%02X%02X\n", chunk_info->info[0], chunk_info->info[1], chunk_info->info[2], chunk_info->info[3]);
                    }
                    printf("		file_pos = %lld (0x%08llX)\n", (long long)chunk_info->file_pos, (long long)chunk_info->file_pos);
                    printf("		pack_pos = %lld (0x%08llX)\n", (long long)chunk_info->pack_pos, (long long)chunk_info->pack_pos);
                    printf("		length = %d\n", chunk_info->length);
                    printf("		aes3_channel = %d\n", chunk_info->aes3_channel);
                    printf("		aes3_payload_length = %d\n", chunk_info->aes3_payload_length);
                    printf("		PTS = %lld\n", (long long)chunk_info->PTS);

                    printf("\n");
                }
            }
            break;

        case SAMPLE_INFO:
            if (stream_info->vars->print_auxinfo) {
                sample_info = (sample_struct*)info_ptr;
                if (sample_info && (info_size == sizeof(sample_struct)))
                {
                    printf("SAMPLE_INFO %d:\n", stream_info->idx);
                    printf("		rtStart = %lld\n", (long long)sample_info->rtStart);
                    printf("		rtStop = %lld\n", (long long)sample_info->rtStop);
                }
            }
            break;

        case TIMECODE_INFO:
            if (stream_info->vars->print_auxinfo) {
                time_code_info = (mc_time_code_t*)info_ptr;
                if (time_code_info && (info_size == sizeof(mc_time_code_t)))
                {
                    printf("TIMECODE_INFO %d:\n", stream_info->idx);
                    printf("		%.02d:%.02d:%.02d:%.02d\n", time_code_info->Hour, time_code_info->Min, time_code_info->Sec, time_code_info->Frm);
                }
            }
            break;
    }

    return BS_OK;
}

static void show_usage()
{
    printf("\nUsage: \n   test_demux_mxf_ext_io  <input filename>  <options> id=filename  id=filename ... \n");
    printf("Where 'id' is the stream ID of the essence stream to demux into 'filename'.\n");
    printf("Run with input filename only to get the list of streams printed.\n\n");
    printf("Options:\n");
    printf("   -p Specify preview seek range, e.g. -p 500.\n");
    printf("   -g Process input file as growing\n");
    printf("   -print_auxinfo Print auxinfo.\n");
    printf("   -use_ext_io Use external IO.\n");
    printf("   -ext_log_file <filename> Specify external log file.\n");
    printf("\n");
}

int32_t parse_args(int argc, char *argv[], char *opt_str, char *opt_value, global_vars_struct& global_vars, char * index_file_path){
    int32_t i(1);
    int32_t num_opts(0);
    char *ptr;
    while (i < argc)
    {
        strcpy(opt_str, argv[i]);
        if (i == 1)
        {
            strcpy(global_vars.input_file, opt_str);
        }
        else if( 0 == strcmp(opt_str, "-p") )
        {
            strcpy(opt_value, argv[++i]);
            global_vars.preview_range = atoi(opt_value);
            num_opts++;
        }
        else if (0 == strcmp(opt_str, "-g")) {
            global_vars.is_growing_file = true;
            num_opts++;
        }
        else if (0 == strcmp(opt_str, "-print_auxinfo")) {
            global_vars.print_auxinfo = true;
            num_opts++;
        }
        else if (0 == strcmp(opt_str, "-use_ext_io"))
        {
            global_vars.use_ext_io = true;
            num_opts++;
        }
        else if (0 == strcmp(opt_str, "-ext_log_file"))
        {
            if (++i >= argc)
            {
                show_usage();
                return -1;
            }
            strcpy(global_vars.ext_log_file, argv[i]);
            num_opts++;
        }
        else
        {
            ptr = strrchr(opt_str, '=');
            if (!ptr)
            {
                show_usage();
                return -1;
            }
            ptr[0] = 0;
            ptr++;

            if (sscanf(opt_str, "%d", &global_vars.streams_to_get[global_vars.num_streams_to_get].ID) != 1)
            {
                show_usage();
                return -1;
            }
            strcpy(global_vars.streams_to_get[global_vars.num_streams_to_get].path, ptr);

            global_vars.num_streams_to_get++;
        }
        i++;
    }
    global_vars.gfile_open_timeout = 0;
    global_vars.dmux_set.format_pcm_output_flag = 1;
    return 0;
}

void cleanup(global_vars_struct& global_vars){
    if (global_vars.parser)
        mxfDemuxFreeParser(global_vars.demuxer, global_vars.parser);
    if (global_vars.demuxer)
        mxfDemuxFree(global_vars.demuxer);

    printf("\n");

    for (int32_t i = 0; i < MAX_STREAMS; i++)
    {
        if (global_vars.streams[i].set.bs)
        {
            int64_t bytes_written = 0;
            global_vars.streams[i].set.bs->auxinfo(global_vars.streams[i].set.bs, 0, BYTECOUNT_INFO, &bytes_written, sizeof(bytes_written));
            printf("Stream ID %d demuxed %lld bytes\n", global_vars.streams[i].stream_format.ID, (long long)global_vars.streams[i].total_demuxed);
            close_file_buf(global_vars.streams[i].set.bs, 0);
        }
    }
    printf("\n");
}

static long long int monotonic_time()
{
#if defined(_WIN32) || defined(_WIN64)
    return static_cast<long long int>(clock() / CLOCKS_PER_SEC);
#elif defined(__APPLE__)
    mach_timebase_info_data_t info;
    mach_timebase_info(&info);
    return static_cast<long long int>(mach_absolute_time() * info.numer / info.denom / 1000000000);
#else
    timespec time;
    clock_gettime(CLOCK_MONOTONIC, &time);
    return static_cast<long long int>(time.tv_sec);
#endif
}

int main(int argc, char *argv[])
{
    global_vars_struct global_vars{};
    int32_t stream_num;
    //int64_t prg_mod, prg_next;
    char opt_str[2048], opt_value[256], index_file_path[512];

    memset(&global_vars, 0, sizeof(global_vars_struct));

    /////////////////////////////////////////////////////////////////
    // check arguments
    if (argc < 2)
    {
        show_usage();
        return 1;
    }

    /////////////////////////////////////////////////////////////////
    // set param defaults
    global_vars.gfile_open_timeout = 0;

    if( parse_args(argc, argv, opt_str, opt_value, global_vars, index_file_path) < 0 ){
        return -1;
    }
    ///////////////////////////////////////////////////////////////////////
    // open input file

    global_vars.dmux_set.parse_external_streams = 1;    // link OP-Atom files
    global_vars.dmux_set.deliver_aes3_331_headers = 1;  // output AES-3 331 audio with element headers
    global_vars.dmux_set.preview_range = global_vars.preview_range;
    global_vars.dmux_set.is_growing_file = global_vars.is_growing_file ? 1 : 0;
    if (global_vars.is_growing_file) {
        // Growing file can be completely empty when opened initially.
        // Give demultiplexer a time, expect at least structural data to appear within this time.
        global_vars.dmux_set.growing_files_open_timeout = global_vars.gfile_open_timeout;
    }
    std::unique_ptr<CExtIO> ext_io;
    if (global_vars.use_ext_io)
    {
        ext_io = std::unique_ptr<CExtIO>{new CExtIO(global_vars.input_file, global_vars.dmux_set, global_vars.ext_log_file)};
    }

    callbacks_t callbacks;
    init_callbacks(callbacks);
    global_vars.demuxer = mxfDemuxNew(&callbacks, NULL, &global_vars.dmux_set);
    if (!global_vars.demuxer)
    {
        printf("Unable create instance of demuxer\n");
        return 1;
    }

    if (mxfDemuxOpen(global_vars.demuxer, NULL, global_vars.input_file))
    {
        printf("Unable to open input file\n");
        cleanup(global_vars);
        return -1;
    }


    ///////////////////////////////////////////////////////////////////////
    // print streams info
    if( detect_contents(&global_vars) ){
        cleanup(global_vars);
        return -1;
    }

    if (global_vars.num_streams_to_get == 0)
    {
        //we're done
        mxfDemuxFree(global_vars.demuxer);
        return 0;
    }

    ///////////////////////////////////////////////////////////////////////////
    // create parser

    global_vars.parser = mxfDemuxNewParser(global_vars.demuxer, &global_vars.parser_set);
    if (!global_vars.parser)
    {
        printf("Unable to create new parser\n");
        cleanup(global_vars);
        return -1;
    }

    stream_num = -1;

    //////////////////////////////////////////////////////////////////////////
    // create output streams
    {
        int32_t j(0);
        for (int32_t i = 0; i < global_vars.num_streams_to_get; i++)
        {
            // find the ID in the stream list
            bool found = false;
            while (j < global_vars.stream_count)
            {
                if (global_vars.streams_to_get[i].ID == global_vars.streams[j].stream_format.ID)
                {
                    found = true;
                    // this is a stream we want
                    global_vars.streams[j].set.stream_num = j;
                    global_vars.streams[j].set.bs = open_file_buf_write(global_vars.streams_to_get[i].path, 10 * 1024 * 1024, NULL);
                    if (global_vars.streams[j].set.bs == nullptr)
                    {
                        printf("Invalid path specified for stream %d\n", global_vars.streams_to_get[i].ID);
                        show_usage();
                        cleanup(global_vars);
                        return -1;
                    }

                    global_vars.streams[j].vars = &global_vars;

                    global_vars.streams[j].set.bs->drive_ptr = (struct drive_struct*)&global_vars.streams[j];

                    // set custom auxinfo handler to retrieve AES-3 331 header, timestamps, timecodes etc..
                    global_vars.streams[j].set.bs->auxinfo = demux_auxinfo;

                    if (mxfDemuxAddStream(global_vars.demuxer, global_vars.parser, &global_vars.streams[j].set))
                    {
                        printf("Unable to add stream %d\n", global_vars.streams_to_get[i].ID);
                        cleanup(global_vars);
                        return -1;
                    }

                    if (stream_num < 0)
                        stream_num = j;

                    break;
                }
                j++;
            }

            if (!found)
            {
                // stream not found!
                printf("Invalid stream ID %d\n", global_vars.streams_to_get[i].ID);
                show_usage();
                cleanup(global_vars);
                return -1;
            }

        }
    }


    mxfdmux_seek_settings_t seek_set;
    memset(&seek_set, 0, sizeof(mxfdmux_seek_settings_t));

    ///////////////////////////////////////////////////////////////////////////////////
    // seek to end - example

    uint64_t _sample_count = global_vars.streams[stream_num].stream_format.positionable_sample_count;

    if (global_vars.streams[stream_num].stream_format.format.stream_mediatype > mctMinPCMType &&
        global_vars.streams[stream_num].stream_format.format.stream_mediatype < mctMaxPCMType)
    {
        printf("Seek to audio stream end of positionable range\n");
        seek_set.seek_flags = SEEKING_FLAG_BY_TIME | SEEKING_FLAG_RETURN_REF_TIME | SEEKING_FLAG_RETURN_TIME;
        // seek by time
        mc_audio_format *audio_fmt = (mc_audio_format*)global_vars.streams[stream_num].stream_format.format.pFormat;

        double clocks_per_sample = (27000000.0 / (double)audio_fmt->samples_per_sec) - 0.1;
        // position to last sample:
        seek_set.start_pos = (int64_t)(double(_sample_count - 1) * clocks_per_sample);
        if (mxfDemuxSetPositions(global_vars.demuxer, global_vars.parser, &seek_set))
        {
            printf("Unable to seek to end of positionable range\n");
        }
    }
    else if (global_vars.streams[stream_num].stream_format.format.stream_mediatype > mctMinVideoType &&
        global_vars.streams[stream_num].stream_format.format.stream_mediatype < mctMaxVideoType)
    {
        printf("Seek to video stream end of positionable range\n");
        // position to last Frame:

        // seek by frame number
        seek_set.seek_flags = SEEKING_FLAG_BY_FRAME | SEEKING_FLAG_RETURN_REF_FRAME | SEEKING_FLAG_RETURN_TIME;
        seek_set.start_pos = _sample_count - 1;

        // alternatively  seek by time
        //mc_video_format *video_fmt = (mc_video_format*)global_vars.streams[stream_num].stream_format.format.pFormat;
        //double fps = 10000000.0 / (double)video_fmt->avg_time_per_frame;
        //clocks_per_sample = 27000000.0 / fps;
        //seek_set.start_pos = (int64_t)(double (sample_count -1) * clocks_per_sample);

        if (mxfDemuxSetPositions(global_vars.demuxer, global_vars.parser, &seek_set))
        {
            printf("Unable to seek to end of positionable range\n");
        }
    }

    /////////////////////////////////////////////////////////////////////////////////////////
    // demultiplex

    printf("\nSeek to start position\n");
    seek_set.seek_flags = SEEKING_FLAG_BY_FRAME | SEEKING_FLAG_RETURN_REF_FRAME | SEEKING_FLAG_RETURN_TIME;
  if( global_vars.seek_position_stop == -1 ){
        global_vars.seek_position_stop = global_vars.streams[stream_num].stream_format.index_num_entries;
    }

  if( global_vars.seek_position_start > global_vars.streams[stream_num].stream_format.index_num_entries ) {
        printf("\nWrong seek position!\n");
        return -1;
    }

    seek_set.start_pos = global_vars.seek_position_start;
    seek_set.stop_pos = global_vars.seek_position_stop;

    int32_t _result = mxfDemuxSetPositions(global_vars.demuxer, global_vars.parser, &seek_set);
  if( _result != MXF_SEEK_SUCCESSFULL ){
        printf("\nUnable to seek on start position\n");
        cleanup(global_vars);
        return -1;
    }

    if (seek_set.seek_flags & (SEEKING_FLAG_BY_TIME | SEEKING_FLAG_RETURN_REF_TIME))
        printf("\nPositioned to ref time %.2f sec, start sending data...\n\n", (double)(seek_set.ref_time) / 27000000.0);
    else if (seek_set.seek_flags & (SEEKING_FLAG_BY_ABSOLUTE_OFFSET | SEEKING_FLAG_RETURN_REF_OFFSET))
        printf("\nPositioned to ref offset 0x%16" PRIx64 ", start sending data...\n\n", seek_set.ref_offset);
    else if (seek_set.seek_flags & (SEEKING_FLAG_BY_FRAME | SEEKING_FLAG_RETURN_REF_FRAME))
        // printf("\nPositioned to ref frame #%I64d, start sending data...\n\n", (long long)seek_set.ref_frame);
        printf("\nPositioned to ref frame #%" PRId64 " start sending data...\n\n", seek_set.ref_frame);

    // now do the demuxing
    int64_t previous_file_size = global_vars.file_info.file_size;
    long long int file_size_update_time = monotonic_time();
    while (true) {
        const int32_t i = mxfDemuxPush(global_vars.demuxer, global_vars.parser);
        if (global_vars.is_growing_file) {
            mxfDemuxGetFileInfo(global_vars.demuxer, &global_vars.file_info);
        }
        if (i > 0) {
            if (global_vars.curr_processed > global_vars.prev_processed) {
                printf("Processed %lld bytes of %lld ...\n", (long long)global_vars.curr_processed, (long long)global_vars.file_info.file_size);
                global_vars.prev_processed = global_vars.curr_processed;
            }
        }
        else if (global_vars.is_growing_file) {
            if (previous_file_size < global_vars.file_info.file_size) {
                // File is still growing ...
                previous_file_size = global_vars.file_info.file_size;
                file_size_update_time = monotonic_time();
                if (global_vars.use_ext_io && ext_io) {
                    mxfDemuxUpdateFilesize(global_vars.demuxer, global_vars.parser, ext_io->GetFileSize(global_vars.input_file));
                }
            }
            else if (monotonic_time() - file_size_update_time > 10) {
                // File is not growing for last 10 seconds
                printf("File is not growing anymore, exiting ...\n");
                break;
            }
            // File is still growing, sleep and continue ...
#if defined(_WIN32) || defined(_WIN64)
            Sleep(100);
#else
            usleep(100 * 1000);
#endif
        }
        else {
            if (i < 0)
                printf("mxfDemuxPush returned negative value\n");
            break;
        }
    }

    cleanup(global_vars);
    return 0;
}
